"""Access the Azure HTTP API"""

from __future__ import annotations

import dataclasses
import json
import logging
import re
import time
from typing import Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union, cast

import requests
from pydantic import BaseModel, TypeAdapter, ValidationError

from llamazure.azrest.models import AzBatch, AzBatchResponse, AzBatchResponses, AzList, AzureError, AzureErrorResponse, BatchReq, Req, Ret_T

l = logging.getLogger(__name__)


K = TypeVar("K")
V = TypeVar("V")


def _partition_dict(d: dict[K, V], p: Callable[[K, V], bool]) -> tuple[dict[K, V], dict[K, V]]:
	"""Partition an iterable into those that match the predicate and those that do not."""
	matches = {}
	nonmatches = {}

	for k, v in d.items():
		if p(k, v):
			matches[k] = v
		else:
			nonmatches[k] = v

	return matches, nonmatches


def _partition_list(l: list[V], p: Callable[[V], bool]) -> tuple[list[V], list[V]]:
	"""Partition an iterable into those that match the predicate and those that do not."""
	matches = []
	nonmatches = []

	for v in l:
		if p(v):
			matches.append(v)
		else:
			nonmatches.append(v)

	return matches, nonmatches


def fmt_req(req: Req) -> str:
	"""Format a request"""
	return req.name


def fmt_log(msg: str, req: Req, **kwargs: Union[str, int, float]) -> str:
	"""Format a log statement referencing a request"""
	arg_s = " ".join(f"{k}={v}" for k, v in kwargs.items())
	return f"{msg} req={fmt_req(req)} {arg_s}"


@dataclasses.dataclass
class RetryPolicy:
	"""Parameters and strategies for retrying Azure Resource Graph queries"""

	retries: int = 0  # number of times to retry. This is in addition to the initial try
	long_running_retries: int = 10  # number of retry attempts to try each long-running task. This is in addition to the initial try
	long_running_wait_multiplier: float = 1.0  # what fraction of the given Retry-After time to wait
	wait_for_ratelimit: bool = True  # whether to wait for ratelimit or return the error
	retry_batch_individually: bool = True  # whether to retry individual members of batches


HEADER_LOCATION = "Location"
HEADER_ASYNC = "Azure-AsyncOperation"
HEADER_RETRY_AFTER = "Retry-After"
HEADER_RATELIMIT_RETRY_AFTER = re.compile("x-ms-ratelimit-.*-retry-after")


class AzRest:
	"""Access the Azure HTTP API"""

	def __init__(self, session: requests.Session, base_url: str = "https://management.azure.com", retry_policy: RetryPolicy = RetryPolicy()):
		self.session = session

		self.base_url = base_url
		self.retry_policy = retry_policy

	@classmethod
	def from_credential(cls, credential, token_scope="https://management.azure.com//.default", base_url="https://management.azure.com") -> AzRest:
		"""Create from an Azure credential"""
		token = credential.get_token(token_scope)
		session = requests.Session()
		session.headers["Authorization"] = f"Bearer {token.token}"

		return cls(session=session, base_url=base_url)

	def to_request(self, req: Req) -> requests.Request:
		"""Convert a Req into a requests.Request"""
		r = requests.Request(method=req.method, url=self.base_url + req.path)
		if req.params:
			r.params = req.params
		if req.apiv:  # TODO: isn't this always required?
			r.params["api-version"] = req.apiv
		if req.body:
			r.headers["Content-Type"] = "application/json"
			if isinstance(req.body, dict):
				# allows you to do your own serialisation
				r.data = json.dumps(req.body)
			else:
				r.data = req.body.model_dump_json(exclude_none=True, by_alias=True)
		return r

	def _build_url(self, req: Req) -> str:
		"""Hacky way to get requests to build our url for us"""
		return cast(str, self.session.prepare_request(self.to_request(req)).url)

	def _to_batchable_request(self, req: Req, batch_id: str) -> Dict[str, Union[str, BaseModel]]:
		r: Dict[str, Union[str, BaseModel]] = {
			"httpMethod": req.method,
			"name": batch_id,
			"url": self._build_url(req),
		}
		if req.body:
			r["content"] = req.body
		return r

	def batch_to_request(self, batch: BatchReq) -> Req[AzBatchResponses]:
		"""Convert the BatchReq into the Req that contains the requests"""
		req = Req(
			name=batch.name,
			path="/batch",
			method="POST",
			apiv=batch.apiv,
			body=AzBatch(requests=[self._to_batchable_request(r, batch_id) for batch_id, r in batch.requests.items()]),
			ret_t=AzBatchResponses,
		)
		return req

	def _resolve_batch_response(self, req: Req[Ret_T], res) -> Union[Ret_T, AzureError]:
		"""Deserialise the response to a batch request"""
		if res.content.get("error"):
			return AzureErrorResponse.model_validate(res.content).error.as_exception()
		type_adapter = TypeAdapter(req.ret_t)
		return type_adapter.validate_python(res.content)

	def call_batch(self, req: BatchReq) -> Dict[str, Union[Ret_T, AzureError]]:
		"""Call a batch request"""
		if self.retry_policy.retry_batch_individually:
			successes, errors = self._call_batch_with_individual_retry(req)
			all_responses = [*successes, *errors]
		else:
			batch_response = self.call(self.batch_to_request(req))
			all_responses = batch_response.responses
		return {e.name: self._resolve_batch_response(req.requests[e.name], e) for e in all_responses}

	def _get_ratelimit_time_to_wait(self, error: AzBatchResponse) -> float:
		"""Get the time to wait from the first plausible retry-after header. Defaults to 0."""
		for header in error.headers:
			if HEADER_RATELIMIT_RETRY_AFTER.fullmatch(header):
				return float(error.headers[header])
		return 0

	def _call_batch_with_individual_retry(self, req: BatchReq, attempt=0) -> Tuple[List[AzBatchResponse], List[AzBatchResponse]]:
		"""Call a batch request with retries for the individual requests."""

		results = self.call(self.batch_to_request(req))

		successes, errors = _partition_list(results.responses, lambda v: v.content is not None and "error" in v.content)

		if errors and attempt < self.retry_policy.retries:
			new_batch = BatchReq({k: e for k, e in req.requests.items() if k in errors})
			time_to_sleep = max(self._get_ratelimit_time_to_wait(e) for e in errors)
			time.sleep(time_to_sleep)

			new_successes, new_errors = self._call_batch_with_individual_retry(new_batch, attempt + 1)
			successes.extend(new_successes)
			errors = new_errors  # either the old errors succeeded or are superseded by these, so we overwrite

		return successes, errors

	def call(self, req: Req[Ret_T]) -> Ret_T:
		"""Make the request to Azure"""
		r = self.to_request(req)
		res = self._deserialise(req, self._call_with_retry(req, r))
		if res is None:
			return res

		if isinstance(res, AzList):
			res_list: AzList = res
			acc = res.value
			page = 0
			while res_list.nextLink:
				page += 1
				l.debug(fmt_log("paginating req", req, page=str(page)))
				# This is basically always a GET
				# TODO: support the nextLink.operationName
				r = requests.Request(method="GET", url=res_list.nextLink)
				res_list = self._deserialise(req, self._call_with_retry(req, r))  # type: ignore  # we know the req
				acc.extend(res_list.value)
			return acc  # type: ignore  # we're deliberately unwrapping a list into its primitive type
		else:
			return res

	def _call_with_retry(self, req: Req[Ret_T], r: requests.Request) -> requests.Response:
		l.debug(fmt_log("making req", req))
		res = self._do_call(r)
		if isinstance(res, AzureError):
			retries = 0
			while retries < self.retry_policy.retries and isinstance(res, AzureError):
				l.debug(fmt_log("req returned error; retrying", req, err=res.error.model_dump_json()))
				retries += 1
				res = self._do_call(r)  # type: ignore

		if isinstance(res, AzureError):
			l.warning(fmt_log("req returned error; retries exhausted", req, err=res.error.model_dump_json()))
			raise res
		else:
			l.debug(fmt_log("req complete", req))
			return res

	def _do_call(self, r: requests.Request) -> Union[requests.Response, AzureError]:
		"""Make a single request to Azure, without retry or pagination"""
		res = self.session.send(self.session.prepare_request(r))
		if not res.ok:
			return AzureErrorResponse.model_validate_json(res.content).error.as_exception()
		return res

	def _deserialise(self, req: Req[Ret_T], res: requests.Response) -> Ret_T:
		if req.ret_t is Type[None]:  # noqa: E721  # we're comparing types here
			return None  # type: ignore

		type_adapter = TypeAdapter(req.ret_t)
		if len(res.content) == 0:
			return type_adapter.validate_python(None)

		deserialised = type_adapter.validate_json(res.content)
		return deserialised

	def _get_longpoll_location(self, res: requests.Response) -> Optional[str]:
		if HEADER_ASYNC in res.headers:
			return res.headers[HEADER_ASYNC]
		elif HEADER_LOCATION in res.headers:
			return res.headers[HEADER_LOCATION]
		else:
			return None

	def _get_longpoll_time_to_wait(self, res: requests.Response) -> float:
		if HEADER_RETRY_AFTER in res.headers:
			return float(res.headers[HEADER_RETRY_AFTER]) * self.retry_policy.long_running_wait_multiplier
		else:
			return 5.0 * self.retry_policy.long_running_wait_multiplier

	def call_long_operation(self, req: Req[Ret_T]) -> Ret_T:
		"""Make a call for a long-running operation, where we will need to check a new location for the result."""
		ir = self._call_with_retry(req, self.to_request(req))

		if ir.status_code not in {202, 201}:
			l.warning(fmt_log("req longpoll returned unexpected status", req, status=ir.status_code))

		result_location = self._get_longpoll_location(ir)
		if result_location is None:
			msg = f"req longpoll did not have header needed to find result of longpoll, expected one of '{HEADER_ASYNC}' or '{HEADER_LOCATION}'"
			l.error(fmt_log(msg, req))
			raise RuntimeError(msg)  # TODO: use real class

		done_req = Req.from_url(req.name, "GET", result_location.removeprefix(self.base_url), ret_t=req.ret_t)  # TODO: better name

		res = self._call_with_retry(done_req, self.to_request(done_req))

		retries = 0
		while retries < self.retry_policy.long_running_retries and res.status_code != 200:
			time_to_wait = self._get_longpoll_time_to_wait(res)
			l.debug(fmt_log("longpoll request sleep", req, attempt=retries, time=time_to_wait))
			time.sleep(time_to_wait)
			res = self._call_with_retry(done_req, self.to_request(done_req))

		if res.status_code not in {200, 204}:
			try:
				error = AzureErrorResponse.model_validate_json(res.content).error.as_exception()
			except ValidationError:
				msg = "req longpoll returned unexpected status that could not be deserialised into an error"
				l.error(fmt_log(msg, req, status=res.status_code, content=res.content.decode()))
				raise RuntimeError(msg)
			raise error
		else:
			return self._deserialise(req, res)


class AzOps:
	"""Parent class for helpers which dispatch requests to Azure"""

	def __init__(self, azrest: AzRest):
		self.azrest = azrest

	def run(self, req: Req[Ret_T]) -> Ret_T:
		"""Call a request"""
		return self.azrest.call(req)


def rid_eq(a: Optional[str], b: Optional[str]) -> bool:
	"""Whether 2 Azure resource IDs are the same"""
	return a is not None and b is not None and a.lower() == b.lower()



import os

import click


def count_occurrences(filepath, search_string):
	try:
		with open(filepath, "r", errors="ignore") as file:
			content = file.read()
			return content.count(search_string)
	except Exception as e:
		print(f"Error reading file {filepath}: {str(e)}")
		return 0


def search_files(directory, search_string, file_occurrences):
	for filename in os.listdir(directory):
		filepath = os.path.join(directory, filename)

		if os.path.isdir(filepath):
			# Recursively search in subdirectories
			file_occurrences = search_files(filepath, search_string, file_occurrences)
		elif os.path.isfile(filepath):
			# Count occurrences of the search string in the file
			occurrences = count_occurrences(filepath, search_string)

			# Update the maximum occurrences for the file
			file_occurrences[filename] = max(file_occurrences.get(filename, 0), occurrences)

	return file_occurrences


@click.command()
@click.option("--start-directory", default=".", help="The directory to start the search.")
@click.option("--search-string", prompt="Enter the search string", help="The string to search for in files.")
def main(start_directory, search_string):
	"""Search for count occurrences of a string, deduplicating by file name"""
	# Perform the search
	file_occurrences = search_files(start_directory, search_string, {})

	# Sort files by number of occurrences
	sorted_files = sorted(file_occurrences.items(), key=lambda x: x[1], reverse=True)

	# Print the sorted files and their occurrences
	print("Files sorted by number of occurrences:")
	for file_name, occurrences in sorted_files:
		if occurrences > 0:
			print(f"{file_name}: {occurrences}")

	# Print the count of files with 0 occurrences
	files_with_zero_occurrences = sum(1 for occurrences in file_occurrences.values() if occurrences == 0)
	print(f"Files with 0 occurrences: {files_with_zero_occurrences}")


if __name__ == "__main__":
	main()  # pylint: disable=no-value-for-param



"""Models for the Azure REST API"""

from __future__ import annotations

import dataclasses
import uuid
from dataclasses import dataclass, field
from typing import Any, Dict, Generic, Iterator, List, Optional, Type, TypeVar, Union
from urllib.parse import parse_qs, urlparse

from pydantic import BaseModel, BeforeValidator, Field

Ret_T = TypeVar("Ret_T")
Ret_T0 = TypeVar("Ret_T0")
ReadOnly = Optional[Ret_T]


# pylint: disable=too-many-arguments
@dataclass(frozen=True)
class Req(Generic[Ret_T]):
	"""Azure REST request"""

	name: str
	path: str
	method: str
	apiv: Optional[str]
	body: Optional[BaseModel] = None
	params: Dict[str, Union[str, List[str]]] = field(default_factory=dict)
	ret_t: Type[Ret_T] = Type[None]  # type: ignore

	@classmethod
	def get(cls, name: str, path: str, apiv: str, ret_t: Type[Ret_T]) -> Req:
		"""Create a GET request"""
		return cls(name, path, "GET", apiv, ret_t=ret_t)

	@classmethod
	def delete(cls, name: str, path: str, apiv: str, ret_t: Union[Any, Optional[Type[Ret_T]]] = Type[None]) -> Req:
		"""Create a DELETE request"""
		return cls(name, path, "DELETE", apiv, ret_t=ret_t)  # type: ignore

	@classmethod
	def put(cls, name: str, path: str, apiv: str, body: Optional[BaseModel], ret_t: Type[Ret_T]) -> Req:
		"""Create a PUT request"""
		return cls(name, path, "PUT", apiv, body, ret_t=ret_t)

	@classmethod
	def post(cls, name: str, path: str, apiv: str, body: Optional[BaseModel], ret_t: Type[Ret_T]) -> Req:
		"""Create a POST request"""
		return cls(name, path, "POST", apiv, body, ret_t=ret_t)

	@classmethod
	def patch(cls, name: str, path: str, apiv: str, body: Optional[BaseModel], ret_t: Type[Ret_T]) -> Req:
		"""Create a PATCH request"""
		return cls(name, path, "PATCH", apiv, body, ret_t=ret_t)

	def named(self, name: str) -> Req:
		"""Change the friendly name of this request"""
		return dataclasses.replace(self, name=name)

	def add_params(self, params: Dict[str, str]) -> Req:
		"""Add query params to this request"""
		return dataclasses.replace(self, params={**self.params, **params})

	def add_param(self, name: str, value: str) -> Req:
		return dataclasses.replace(self, params={**self.params, **{name: value}})

	def with_ret_t(self, ret_t: Type[Ret_T0]) -> Req[Ret_T0]:
		"""Override the return type"""
		return dataclasses.replace(self, ret_t=ret_t)  # type: ignore

	@classmethod
	def from_url(
		cls,
		name: str,
		method: str,
		url: str,
		body: Optional[BaseModel] = None,
		ret_t: Type[Ret_T] = Type[None],  # type: ignore
	) -> Req:
		"""
		Build a Req from a URL, useful for when Azure has handed you one; for example, from the 'Azure-AsyncOperation' header.

		This function does less validation, on the assumption Azure knows what we should do.
		"""
		parsed = urlparse(url)

		path = parsed.path
		params = parse_qs(parsed.query)
		built_params = {k: v[0] if len(v) == 1 else v for k, v in params.items()}
		apiv = built_params.pop("api-version", None)  # type: ignore # this shouldn't happen but there's no reason for us to throw

		return Req(
			name=name,
			path=path,
			method=method,
			apiv=apiv,  # type: ignore # we know what we're doing
			body=body,
			params=built_params,
			ret_t=ret_t,
		)


@dataclass
class BatchReq:
	"""A batch of requests to the Azure API"""

	requests: Dict[str, Req]
	name: str = "batch"
	apiv: str = "2020-06-01"

	@classmethod
	def gather(cls, reqs: List[Req], name: str = "batch", apiv: str = "2020-06-01") -> BatchReq:
		"""Gather many requests into a batch, automatically assigning them an ID"""
		keyed_requests = {str(uuid.uuid4()): r for r in reqs}
		return cls(keyed_requests, name, apiv)


class AzBatch(BaseModel, Generic[Ret_T]):
	"""A serialisable request to the batch API"""

	requests: List[Ret_T]


class AzBatchResponse(BaseModel):
	"""A single response in a batch"""

	name: str
	httpStatusCode: int
	headers: Dict[str, str] = {}
	content: Optional[Dict]


class AzBatchResponses(BaseModel):
	"""The bundle of responses from the Azure batch API"""

	responses: List[AzBatchResponse]


class AzList(BaseModel, Generic[Ret_T]):
	"""A deserialisation of a List from Azure"""

	value: List[Ret_T]
	nextLink: Optional[str] = None

	def __iter__(self) -> Iterator[Ret_T]:  # type: ignore[override]
		# TODO: explode if there's a nextLink because this won't paginate properly
		return self.value.__iter__()


default_list = BeforeValidator(lambda v: v if v is not None else [])
default_dict = BeforeValidator(lambda v: v if v is not None else {})


class AzureError(Exception):
	"""An error from the Azure API"""

	def __init__(self, error: AzureErrorDetails):
		super().__init__(error.message)  # TODO: better stringify message
		self.error = error


class AzureErrorResponse(BaseModel):
	"""The container of an Azure error"""

	error: AzureErrorDetails


class AzureErrorDetails(BaseModel):
	"""An Azure-specific error"""

	code: str
	message: str
	target: Optional[str] = None
	details: List[AzureErrorDetails] = []
	additionalInfo: List[AzureErrorAdditionInfo] = []

	def as_exception(self) -> AzureError:
		"""Wrap this in a Python Exception for throwing"""
		return AzureError(self)


class AzureErrorAdditionInfo(BaseModel):
	"""The resource management error additional info."""

	info_type: str = Field(alias="type")
	info: Dict = {}


T = TypeVar("T")


def ensure(a: Optional[T]) -> T:
	"""Ensure the result is not None"""
	if a is None:
		raise TypeError("value was None")
	return a


P0 = TypeVar("P0", bound=BaseModel)
P1 = TypeVar("P1", bound=BaseModel)


def cast_as(obj: P0, cls: Type[P1]) -> P1:
	"""
	Cast one model into another.

	Useful for turning a Foo into a FooUpdateParameters.
	"""
	return cls.model_validate(obj.model_dump())



"""Inspector to help explore Azure's OpenAPI documents"""

import json
import os
from pathlib import Path
from typing import Any, Callable, Generator, Tuple


def query_files(root_dir: str, callable_func: Callable[[dict], Generator[Any, None, None]]) -> Generator[Tuple[str, Any], None, None]:
	"""
	Walk through a directory tree, open every `.json` file, deserialize it,
	check if it has the key "swagger" set to "2.0", and yield a tuple of the file path
	and the result of invoking a generator function on the document.

	Parameters:
	root_dir (str): The root directory to start walking from.
	callable_func (Callable[[dict], Generator[Any, None, None]]): The generator function to be called on each JSON document that meets the criteria.

	Yields:
	Tuple[str, Any]: A tuple of the file path and the result from the generator function.
	"""
	bad_files = {"tsconfig.json", "devcontainer.json", "settings.json"}

	for dirpath, dirnames, filenames in os.walk(root_dir):
		if "examples" in dirpath:
			continue
		for filename in filenames:
			if filename.endswith(".json") and filename not in bad_files:
				file_path = os.path.join(dirpath, filename)
				try:
					try:
						with open(file_path, "r", encoding="utf-8") as file:
							yield from process_content(callable_func, file, file_path)
					except UnicodeDecodeError:
						with open(file_path, "r", encoding="latin-1") as file:
							yield from process_content(callable_func, file, file_path)
				except Exception:
					print(f"error processing {file_path}")
					raise


def process_content(callable_func, file, file_path):
	try:
		document = json.load(file)
		if document.get("swagger") == "2.0":
			try:
				for result in callable_func(document):
					yield (file_path, result)
			except (TypeError, AttributeError) as e:
				raise RuntimeError(f"error processing {file_path=}") from e
	except (json.JSONDecodeError, IOError) as e:
		raise RuntimeError(f"Error processing file {file_path}") from e


def find_ref_responses(openapi: dict) -> Generator[Path, None, None]:
	for p_name, p in openapi["paths"].items():
		for method, op in p.items():
			if method == "parameters":
				continue
			for code, response in op["responses"].items():
				if "$ref" in response:
					yield Path(op["operationId"], code)


def find_inline_responses(openapi: dict) -> Generator[Path, None, None]:
	for p_name, p in openapi["paths"].items():
		for method, op in p.items():
			if method == "parameters":
				continue
			for code, response in op["responses"].items():
				if "schema" in response:
					schema = response["schema"]
					if "$ref" not in schema:
						if len(schema) > 1:  # if there's just a key, it's probably just `"type": "object"`
							yield Path(op["operationId"], code)


if __name__ == "__main__":
	for v in query_files("/home/lilatomic/vnd/azrest/specification", find_ref_responses):
		print(v)
